import torch
import torch.nn as nn
import torch.nn.functional as F
def conv(in_planes, out_planes, kernel_size=3, stride=1, padding=1, dilation=1):
    return nn.Sequential(
            nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride,
                      padding=padding, dilation=dilation, bias=True),
            nn.BatchNorm2d(out_planes),
            nn.ReLU(inplace=True))

class Mask_Predictor(nn.Module):
    """ Mask Predictor module"""
    def __init__(self, inplanes=64, channel=256):
        super(Mask_Predictor, self).__init__()
        self.conv1 = conv(inplanes,channel)
        self.conv2 = conv(channel,channel)
        self.conv3 = conv(channel,channel)
        self.conv4 = nn.Sequential(nn.Conv2d(channel,1,kernel_size=3,stride=1,padding=1,dilation=1,bias=True),
                                   nn.Sigmoid())
    def forward(self, x):
        """ Forward pass with input x. """
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.conv4(x)
        output = nn.functional.interpolate(x, scale_factor=(16,16), mode='bilinear')#由于下采样了16倍,这里再上采样16倍以与输入分辨率对齐
        return output

class Mask_Predictor_fine_trans(nn.Module):
    def __init__(self):
        super(Mask_Predictor_fine_trans, self).__init__()
        self.v0 = nn.Sequential(
                nn.Conv2d(128, 64, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(64, 16, 3, padding=1),
                nn.ReLU(inplace=True),
            )
        self.v1 = nn.Sequential(
                nn.Conv2d(256, 128, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(128, 64, 3, padding=1),
                nn.ReLU(inplace=True),
            )
        self.v2 = nn.Sequential(
                nn.Conv2d(512, 256, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(256, 256, 3, padding=1),
                nn.ReLU(inplace=True),
            )
        self.h2 = nn.Sequential(
                nn.Conv2d(512, 256, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(256, 256, 3, padding=1),
                nn.ReLU(inplace=True),
            )
        self.h1 = nn.Sequential(
                nn.Conv2d(128, 64, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(64, 64, 3, padding=1),
                nn.ReLU(inplace=True),
            )
        self.h0 = nn.Sequential(
                nn.Conv2d(16, 16, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(16, 16, 3, padding=1),
                nn.ReLU(inplace=True),
            )

        self.post0 = nn.Conv2d(256, 128, 3, padding=1)
        self.post1 = nn.Conv2d(64, 16, 3, padding=1)
        self.post2 = nn.Conv2d(16, 4, 3, padding=1)
        #self.corr_re = nn.Conv2d(512, 256, 3, padding=1)
        self.post_f = nn.Conv2d(4, 1, 3, padding=1)

        ### confidence ###
        self.conv1_cf = conv(512, 256)
        self.conv2_cf = conv(256, 128)
        self.conv3_cf = conv(128, 64)
        self.conv4_cf = conv(64, 32)
        self.conv5_cf = nn.Conv2d(32, 1, kernel_size=1)

        with torch.no_grad():
            self.output_sz = 14  # same as input resolution
            self.indice = torch.arange(0, self.output_sz).view(-1, 1) * 16
            # generate mesh-grid
            self.coord_x = self.indice.repeat((self.output_sz, 1)) \
                .view((self.output_sz * self.output_sz,)).float().cuda()
            self.coord_y = self.indice.repeat((1, self.output_sz)) \
                .view((self.output_sz * self.output_sz,)).float().cuda()

    def get_score_map(self, x):
        x_tl1 = self.conv1_cf(x)
        x_tl2 = self.conv2_cf(x_tl1)
        x_tl3 = self.conv3_cf(x_tl2)
        x_tl4 = self.conv4_cf(x_tl3)
        score_map = self.conv5_cf(x_tl4)
        return score_map # [1,1,14,14]

    def soft_argmax(self, score_map):
        """ get soft-argmax coordinate for a given heatmap """
        prob_vec = nn.functional.softmax(
            score_map.view((-1, self.output_sz * self.output_sz)), dim=1)  # (batch, output_sz * output_sz)
        exp_x = torch.sum((self.coord_x * prob_vec), dim=1)
        exp_y = torch.sum((self.coord_y * prob_vec), dim=1)
        return exp_x, exp_y

    def forward(self, corr_feat, Lfeat):
        """
        corr[b, 512, 14, 14]   128 32
        Lfeat: (batch,128,56,56), (batch,256,28, 28), (batch,512,14,14)'''
        """

        bz, wh, c = corr_feat.shape
        h = int(wh ** 0.5)
        corr_feat_square = corr_feat.reshape(bz, h, -1, c).permute(0, 3, 1, 2)
        for i in range(len(Lfeat)):
            bz, wh, c = Lfeat[i].shape
            h = int(wh ** 0.5)
            Lfeat[i] = Lfeat[i].reshape(bz, h, -1, c).permute(0, 3, 1, 2)

        corr_feat_r = corr_feat_square# self.corr_re(corr_feat_square)
        out = self.post0(F.interpolate(self.h2(corr_feat_r),size=(14,14),mode='bilinear')+self.v2(Lfeat[2])) # (b,32,32,32)+(b,32,32,32) --> (b,16,32,32)
        out = self.post1(F.interpolate(self.h1(out),size=(28,28),mode='bilinear')+self.v1(Lfeat[1])) # (b,16,64,64)+(b,16,64,64) --> (b,4,64,64)
        out = self.post2((F.interpolate(self.h0(out), size=(56, 56),mode='bilinear') + self.v0(Lfeat[0])).contiguous())
        out = self.post_f(out)
        mask = torch.sigmoid(F.interpolate(out, size=(224,224),mode='bilinear')) # (b,1,256,256)

        ###confidence
        score_map = self.get_score_map(corr_feat_square) #[b,1,14,14]
        coorx, coory = self.soft_argmax(score_map)  # t[1],t[1]
        center = torch.stack((coorx, coory), dim=1) # [b, 2]
        return mask, center, score_map

class Mask_Predictor_fine(nn.Module):
    def __init__(self):
        super(Mask_Predictor_fine, self).__init__()
        self.v0 = nn.Sequential(
                nn.Conv2d(64, 16, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(16, 4, 3, padding=1),
                nn.ReLU(inplace=True),
            )
        self.v1 = nn.Sequential(
                nn.Conv2d(256, 64, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(64, 16, 3, padding=1),
                nn.ReLU(inplace=True),
            )
        self.v2 = nn.Sequential(
                nn.Conv2d(512, 128, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(128, 32, 3, padding=1),
                nn.ReLU(inplace=True),
            )
        self.h2 = nn.Sequential(
                nn.Conv2d(64, 32, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(32, 32, 3, padding=1),
                nn.ReLU(inplace=True),
            )
        self.h1 = nn.Sequential(
                nn.Conv2d(16, 16, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(16, 16, 3, padding=1),
                nn.ReLU(inplace=True),
            )
        self.h0 = nn.Sequential(
                nn.Conv2d(4, 4, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(4, 4, 3, padding=1),
                nn.ReLU(inplace=True),
            )

        self.post0 = nn.Conv2d(32, 16, 3, padding=1)
        self.post1 = nn.Conv2d(16, 4, 3, padding=1)
        self.post2 = nn.Conv2d(4, 1, 3, padding=1)

    # def forward(self, corr_feat, Lfeat):
    #     '''corr_feat是经过correlation和Non-local处理后的特征, Lfeat是backbone提取出的底层特征'''
    #     '''corr_feat: (batch,64,16,16)
    #        Lfeat: (batch,64,128,128), (batch,256,64,64), (batch,512,32,32)'''
    #     out = self.post0(F.upsample(self.h2(corr_feat),size=(32,32),mode='bilinear')+self.v2(Lfeat[2])) # (b,32,32,32)+(b,32,32,32) --> (b,16,32,32)
    #     out = self.post1(F.upsample(self.h1(out),size=(64,64),mode='bilinear')+self.v1(Lfeat[1])) # (b,16,64,64)+(b,16,64,64) --> (b,4,64,64)
    #     out = self.post2(F.upsample(self.h0(out), size=(128, 128),mode='bilinear') + self.v0(Lfeat[0]))
    #     out = torch.sigmoid(F.upsample(out,size=(256,256),mode='bilinear')) # (b,1,256,256)
    #     return out
    def forward(self, corr_feat, Lfeat):
        '''corr_feat是经过correlation和Non-local处理后的特征, Lfeat是backbone提取出的底层特征'''
        '''corr_feat: (batch,64,16,16)
           Lfeat: (batch,64,128,128), (batch,256,64,64), (batch,512,32,32)'''
        out = self.post0(F.interpolate(self.h2(corr_feat),size=(32,32),mode='bilinear')+self.v2(Lfeat[2])) # (b,32,32,32)+(b,32,32,32) --> (b,16,32,32)
        out = self.post1(F.interpolate(self.h1(out),size=(64,64),mode='bilinear')+self.v1(Lfeat[1])) # (b,16,64,64)+(b,16,64,64) --> (b,4,64,64)
        out = self.post2((F.interpolate(self.h0(out), size=(128, 128),mode='bilinear') + self.v0(Lfeat[0])).contiguous())
        out = torch.sigmoid(F.interpolate(out,size=(256,256),mode='bilinear')) # (b,1,256,256)
        return out

class Mask_Predictor_mask(nn.Module):
    def __init__(self):
        super(Mask_Predictor_mask, self).__init__()
        self.v0 = nn.Sequential(
                nn.Conv2d(128, 64, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(64, 16, 3, padding=1),
                nn.ReLU(inplace=True),
            )
        self.v1 = nn.Sequential(
                nn.Conv2d(256, 128, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(128, 64, 3, padding=1),
                nn.ReLU(inplace=True),
            )
        self.v2 = nn.Sequential(
                nn.Conv2d(512, 256, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(256, 256, 3, padding=1),
                nn.ReLU(inplace=True),
            )
        self.h2 = nn.Sequential(
                nn.Conv2d(512, 256, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(256, 256, 3, padding=1),
                nn.ReLU(inplace=True),
            )
        self.h1 = nn.Sequential(
                nn.Conv2d(128, 64, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(64, 64, 3, padding=1),
                nn.ReLU(inplace=True),
            )
        self.h0 = nn.Sequential(
                nn.Conv2d(16, 16, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(16, 16, 3, padding=1),
                nn.ReLU(inplace=True),
            )

        self.post0 = nn.Conv2d(256, 128, 3, padding=1)
        self.post1 = nn.Conv2d(64, 16, 3, padding=1)
        self.post2 = nn.Conv2d(16, 4, 3, padding=1)
        #self.corr_re = nn.Conv2d(512, 256, 3, padding=1)
        self.post_f = nn.Conv2d(4, 1, 3, padding=1)


    def forward(self, corr_feat, Lfeat):
        """
        corr[b, 512, 14, 14]   128 32
        Lfeat: (batch,128,56,56), (batch,256,28, 28), (batch,512,14,14)'''
        """
        bz, wh, c = corr_feat.shape
        h = int(wh ** 0.5)
        corr_feat_square = corr_feat.reshape(bz, h, -1, c).permute(0, 3, 1, 2)
        for i in range(len(Lfeat)):
            bz, wh, c = Lfeat[i].shape
            h = int(wh ** 0.5)
            Lfeat[i] = Lfeat[i].reshape(bz, h, -1, c).permute(0, 3, 1, 2)

        corr_feat_r = corr_feat_square# self.corr_re(corr_feat_square)
        out = self.post0(F.interpolate(self.h2(corr_feat_r),size=(14,14),mode='bilinear')+self.v2(Lfeat[2])) # (b,32,32,32)+(b,32,32,32) --> (b,16,32,32)
        out = self.post1(F.interpolate(self.h1(out),size=(28,28),mode='bilinear')+self.v1(Lfeat[1])) # (b,16,64,64)+(b,16,64,64) --> (b,4,64,64)
        out = self.post2((F.interpolate(self.h0(out), size=(56, 56),mode='bilinear') + self.v0(Lfeat[0])).contiguous())
        out = self.post_f(out)
        mask = torch.sigmoid(F.interpolate(out, size=(224,224),mode='bilinear')) # (b,1,256,256)
        return mask


#a= Mask_Predictor_fine()
#a.cuda()

#corr_feat = torch.rand([4, 196, 512]).cuda()

#b1 = torch.rand([4, 56*56, 128]).cuda()
#b2 = torch.rand([4, 28*28, 256]).cuda()
#b3 = torch.rand([4, 14*14, 512]).cuda()

#L=[b1, b2, b3]


#cc= a(corr_feat, L)\
#import  numpy as np

#a= np.array([0,1,2,3])
#b=1